# Template for the Scribe Journal Consumer
SERVICE_TEMPLATE = """[Unit]
Description=Flux Scribe Journal Consumer
After=network.target

[Service]
ExecStart={python_path} -m flux_batch.service.scribe
Restart=on-failure

[Install]
WantedBy=default.target
"""

START_MODULE_TEMPLATE = """
from flux.modprobe import task
import flux.subprocess as subprocess

@task(
    "start-{service_name}",
    ranks="0",
    needs_config=["{service_name}"],
    after=["resource", "job-list"],
)
def start_{service_func}(context):
    # This triggers the systemd user service provisioned earlier
    # context.bash("systemctl --user start {service_name}")
    subprocess.rexec_bg(
        context.handle,
        ["{python_bin}", "-m", "{module_name}"],
        label="{service_name}",
        nodeid=0
    )
"""

STOP_MODULE_TEMPLATE = """
from flux.modprobe import task
import flux.subprocess as subprocess

@task(
    "stop-{service_name}",
    ranks="0",
    needs_config=["{service_name}"],
    before=["resource", "job-list"],
)
def stop_{service_func}(context):
    # context.bash("systemctl --user stop {service_name}")
    subprocess.kill(context.handle, signum=2, label="{service_name}").get()
    try:
        status = subprocess.wait(context.handle, label="{service_name}").get()["status"]
        print(status)
    except:
        pass
"""



import time
from typing import Any, Dict, List, Optional

from sqlalchemy import and_, create_engine, select, update
from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine
from sqlalchemy.orm import sessionmaker

from flux_batch.service.scribe.models import Base, EventModel, EventRecord, JobModel, JobRecord


def _record_event_internal(session, cluster: str, event: Dict[str, Any]):
    """
    Shared synchronous logic for recording events.
    Used by both Sync and Async backends.
    """
    job_id = event.get("id")
    event_type = event.get("type")
    data = event.get("payload", {})
    timestamp = event.get("timestamp", time.time())

    # Add the new event with all metadata
    new_event = EventModel(
        job_id=job_id,
        cluster=cluster,
        timestamp=timestamp,
        event_type=event_type,
        payload=data,
    )
    session.add(new_event)

    if event_type == "submit":
        stmt = select(JobModel).where(and_(JobModel.job_id == job_id, JobModel.cluster == cluster))
        job = session.execute(stmt).scalar_one_or_none()

        if not job:
            job = JobModel(
                job_id=job_id,
                cluster=cluster,
                user=str(data.get("userid", "unknown")),
                state="submitted",
                workdir=data.get("cwd", ""),
                submit_time=timestamp,
                last_updated=timestamp,
            )
            session.add(job)
        else:
            job.state = "submitted"
            job.last_updated = timestamp

    # state transitions
    elif event_type == "state" or (event_type and event_type.endswith(".finish")):
        state_name = data.get("state_name", event_type)
        stmt = select(JobModel).where(and_(JobModel.job_id == job_id, JobModel.cluster == cluster))
        job = session.execute(stmt).scalar_one_or_none()
        if job:
            job.state = state_name
            job.last_updated = time.time()
            if "status" in data:
                job.exit_code = data["status"]


class AsyncSQLAlchemyBackend:
    """
    Asynchronous backend for the MCP Gateway.
    """

    def __init__(self, db_url: str):
        self.engine = create_async_engine(db_url, echo=False)
        self.SessionLocal = async_sessionmaker(self.engine, expire_on_commit=False)

    async def initialize(self):
        async with self.engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

    async def close(self):
        await self.engine.dispose()

    async def record_event(self, cluster: str, event: Dict[str, Any]):
        async with self.SessionLocal() as session:
            # run_sync bridges our shared logic into the async session
            await session.run_sync(_record_event_internal, cluster, event)
            await session.commit()

    async def get_job(self, cluster: str, job_id: int) -> Optional[JobRecord]:
        async with self.SessionLocal() as session:
            result = await session.execute(
                select(JobModel).where(and_(JobModel.job_id == job_id, JobModel.cluster == cluster))
            )
            job = result.scalar_one_or_none()
            return job.to_record() if job else None

    async def get_event_history(self, cluster: str, job_id: int) -> List[EventRecord]:
        async with self.SessionLocal() as session:
            result = await session.execute(
                select(EventModel)
                .where(and_(EventModel.job_id == job_id, EventModel.cluster == cluster))
                .order_by(EventModel.timestamp.asc())
            )
            return [e.to_record() for e in result.scalars().all()]

    async def search_jobs(
        self, cluster: str = None, state: str = None, limit: int = 10
    ) -> List[JobRecord]:
        async with self.SessionLocal() as session:
            stmt = select(JobModel)
            if cluster:
                stmt = stmt.where(JobModel.cluster == cluster)
            if state:
                stmt = stmt.where(JobModel.state == state)
            result = await session.execute(stmt.limit(limit))
            return [j.to_record() for j in result.scalars().all()]


class SQLAlchemyBackend:
    """
    Synchronous backend for the standalone Scribe daemon.
    """

    def __init__(self, db_url: str):
        # strip 'aiosqlite+' or similar if passed from shared config
        url = db_url.replace("+aiosqlite", "").replace("+asyncpg", "")
        self.engine = create_engine(url, echo=False)
        self.SessionLocal = sessionmaker(bind=self.engine, expire_on_commit=False)

    def initialize(self):
        Base.metadata.create_all(self.engine)

    def close(self):
        self.engine.dispose()

    def record_event(self, cluster: str, event: Dict[str, Any]):
        with self.SessionLocal() as session:
            with session.begin():
                _record_event_internal(session, cluster, event)

    def get_unwatched_job_ids(self, cluster: str) -> List[int]:
        """Specific for Scribe: find jobs that need a watcher."""
        with self.SessionLocal() as session:
            stmt = select(JobModel.job_id).where(
                and_(JobModel.cluster == cluster, JobModel.state == "submitted")
            )
            return list(session.execute(stmt).scalars().all())

    def mark_job_as_watched(self, cluster: str, job_id: int):
        with self.SessionLocal() as session:
            with session.begin():
                session.execute(
                    update(JobModel)
                    .where(and_(JobModel.job_id == job_id, JobModel.cluster == cluster))
                    .values(state="watching")
                )



import os
import subprocess
import sys

import flux_batch.service.scribe as scribe

# Lookup of known services
services = {"flux-scribe": scribe.SERVICE_TEMPLATE}
modules = {
    "flux-scribe": {
        "startup": scribe.START_MODULE_TEMPLATE,
        "shutdown": scribe.STOP_MODULE_TEMPLATE,
        "module": scribe.MODULE_NAME,
    }
}


def write_modprobe_script(rc_path, script, args=None):
    """
    Shared function to write service file.
    """
    args = args or {}
    if not os.path.exists(rc_path):
        with open(rc_path, "w") as f:
            f.write(script.format(**args))


def ensure_modprobe_scripts(service_name: str):
    """
    Ensures rc1.d (start) and rc3.d (stop) scripts exist for the service.
    """
    if service_name not in modules:
        print("Warning: module {service_name} is not known.")
        return

    # We will add these to FLUX_MODPROBE_PATH_APPEND
    base_dir = os.path.expanduser("~/.flux-batch")
    for subdir in ["rc1.d", "rc3.d"]:
        os.makedirs(os.path.join(base_dir, subdir), exist_ok=True)

    service_func = service_name.replace("-", "_")

    # Path for rc1.d (startup)
    args = {
        "service_name": service_name,
        "service_func": service_func,
        "python_bin": sys.executable,
        "module_name": modules[service_name]["module"],
    }
    rc1_path = os.path.join(base_dir, "rc1.d", f"{service_name}.py")
    script = modules[service_name]["startup"]
    write_modprobe_script(rc1_path, script, args=args)

    # Path for rc3.d (shutdown)
    args = {"service_name": service_name, "service_func": service_func}
    rc3_path = os.path.join(base_dir, "rc3.d", f"{service_name}.py")
    script = modules[service_name]["shutdown"]
    write_modprobe_script(rc3_path, script, args=args)


def ensure_user_service(service_name: str):
    """
    Checks for the existence of a systemd service file in the user's home.
    If it doesn't exist, it creates it and reloads the daemon.
    """
    user_systemd_dir = os.path.expanduser("~/.config/systemd/user")
    os.makedirs(user_systemd_dir, exist_ok=True)
    service_path = os.path.join(user_systemd_dir, f"{service_name}.service")

    if not os.path.exists(service_path):
        if service_name in services:
            template = services[service_name]
            print(f"[*] Provisioning {service_name} at {service_path}")
            with open(service_path, "w") as f:
                f.write(template.format(python_path=sys.executable))

        else:
            print(f"[*] Service {service_name} is not known, assuming exists.")

        # Reload the user-session manager to recognize the new unit
        subprocess.run(["systemctl", "--user", "daemon-reload"], check=True)
